source("setpar.R")
source("subsample.R")
source("stochastic_simulate1_ctsmr.R")
##
library(ctsmr)
library(MASS)


#set.seed(124)
## Set parameter for an discrete process
n <- 100
prms <- c(X0=0,
          a=0.8,
          b=0.1,
          sigmasys=0.01,
          sigmaobs=0.005)
## Discrete 1,1 process, with a step input
simulateDiscrete <- function(n,x0,a,b,sigmasys,sigmaobs){
    u <- c(rep(0,n/2),rep(0.2,n/2))
    y <- rep(x0,n)
    noise <- rnorm(n)
    for(i in 2:n){
        y[i] <- a*y[i-1] + b*(u[i-1]+u[i])/2 + sigmasys*noise[i]
    }
    z <- y + rnorm(length(y), sd = sigmaobs)
    D <- data.frame(t=1:n, y=z, u=u)
}
D <- simulateDiscrete(n,prms["X0"],prms["a"],prms["b"],prms["sigmasys"],prms["sigmaobs"])
setpar("ts", mfrow=c(5,1))
plot(D$u, type="b", ylab="Simulated sample")
title(main="Simulated sample", line=-1.5)
lines(D$y, type="b")

## Use the discrete simulation to generate many scenarios
## Set how many scenarios
nsim <- 10
ylim <- c(-0.05,0.2)
##
plot(D$t, D$y, type='l', lwd=2, ylim=ylim)
title(main="Scenarios sampled from the population", line=-1.5)
LsimArmax <- lapply(1:nsim, function(i){
    tmp <- simulateDiscrete(n,prms["X0"],prms["a"],prms["b"],prms["sigmasys"],prms["sigmaobs"])
    lines(tmp$t, tmp$y, type='l', col=i)
    return(tmp)
})

## Generate a new object of class ctsm
model <- ctsm()
## Add a system equation and thereby also a state
model$addSystem(dX ~ ( a*X + b*u )*dt + exp(p11)*dw1)
##
model$addInput(u)
## Set the observation equation
model$addObs(y ~ X)
## Set the variance of the measurement error
model$setVariance(y ~ exp(e11))
## Set the initial value (for the optimization) of the value of the state at the starting   time point
model$setParameter(  X0 = c(init=0.1  ,lb=-10     ,ub=10 ) )
## Set the initial value for the optimization
model$setParameter(  a = c(init=0   ,lb=-1  ,ub=1 ) )
model$setParameter(  b = c(init=0   ,lb=-1  ,ub=1 ) )
model$setParameter( p11 = c(init=1  ,lb=-20   ,ub=10 ) )
model$setParameter( e11 = c(init=1  ,lb=-20   ,ub=10 ) )
##


## Run the parameter optimization
fit <- model$estimate(D, firstorder=TRUE)
summary(fit, extended=TRUE)


## The "simulation", well, maybe rather multistep prediction of the mean
sim <- simulate(fit)
## Plot the simulated output
plot(D$t, sim[[1]]$output$sim$y, type='l', ylim=ylim, ylab="Multi-step pred.", col=2)
title("Prediction of the mean (with the simulation function)", line=-1.5)
lines(D$t, D$y, type='l', lwd=1)
lines(D$t, D$u, type='l', lwd=1)


## ## Use the stochastic simulate function from the package
## sim3 <- stochastic.simulate(fit, x0=fit$xm[paste0(model$states,"0")], data=D, dt=0.01)
## ## Plot the simulated output
## plot(D$t[-1], sim3$state$X, type='l', ylab="")
## plot(D$t[-1], sim3$output$y, type='l', col=2)
## ## Looks like it doesn't recognize the input u!?


## Use another the stochastic simulation (made by a hacker (see the original meaning of the word: https://techterms.com/definition/hacker)
plot(D$t, D$y, type='l', lwd=2, ylim=ylim)
title("Output prediction using predict", line=-1.5)
LsimCtsmrEstimated <- lapply(1:nsim, function(i){
    tmp <- stochastic.simulate1.ctsmr(fit, newdata=D, nsub=0, firstorder=TRUE, returnsubsampled=TRUE)
    lines(tmp[[1]]$t, tmp[[1]]$output$y, type='l', col=i)
    return(tmp)
})



############################################################################
## Begin jkmo
t <- seq(min(D$t),max(D$t),by=0.5)
u.fun <- approxfun(D$t,y=D$u)
plot(t,u.fun(t),type="l", ylim=ylim)
title("State prediction using an Euler scheme", line=-1.5)
x0 <- 0
x <- numeric(length(t))
x[1] <- x0
i <- 1
##set.seed(3251)
## for(ii in 1:nsim){
##     for(i in 1:(length(t)-1)){
##         new.dat <- data.frame(t=t[i:(i+1)],y=NA,u=u.fun(t[i:(i+1)]))
##         tmp <- predict(fit, covariance=FALSE, newdata=new.dat,x0=c(X=x[i]),vx0=matrix(1e-32))
##         x[i+1] <- tmp$state$pred$X[2] + rnorm(1,sd=tmp$state$sd$X[2])
##     }
##     lines(t,x,type="l", col=ii)
## }
x.eul <- numeric(length(t))
x.eul[1] <- x0
for(ii in 1:nsim){
    for(i in 1:(length(t)-1)){
        dt <- diff(t[i:(i+1)])
        x.eul[i+1] <- x.eul[i] + (fit$xm["a"]*x.eul[i]+fit$xm["b"]*u.fun(t[i+1]))*dt + exp(fit$xm["p11"]) * rnorm(1)*sqrt(dt)
    }
    lines(t,x.eul,type="l", col=ii)
}

#######################################################################################################
## End jkmo
#######################################################################################################




#######################################################################################################
dev.new()
## Compare with the population values
prmshat <- sapply(1:10, function(i){
    print(i)
    D <- simulateDiscrete(n,prms["X0"],prms["a"],prms["b"],prms["sigmasys"],prms["sigmaobs"])
    ## Run the parameter optimization
    fit <- model$estimate(D, firstorder=TRUE)
    ##    summary(fit, extended=TRUE)
    prmhat <- c(X0 = fit$xm["X0"],
                a = 1+fit$xm["a"],
                b = fit$xm["b"],
                sigmasys = exp(fit$xm["p11"]),
                sigmaobs = sqrt(exp(fit$xm["e11"])))
    return(prmhat)
})
prmshat <- t(prmshat)
##
par(mfrow=c(1,5))
for(i in 1:length(prms)){
    boxplot(prmshat[ ,i], ylim=range(c(prmshat[ ,i],prms[i])))
    points(prms[i], col=2, pch=19, cex=2)
}



setpar("ts", mfrow=c(2,1))
##
## Use the discrete simulation to generate many scenarios
## Set how many scenarios
nsim <- 10
##
plot(D$t, D$y, type='l', lwd=2, ylim=ylim)
title(main="Scenarios sampled from the population", line=-1.5)
LsimArmax <- lapply(1:nsim, function(i){
    tmp <- simulateDiscrete(n,prms["X0"],prms["a"],prms["b"],prms["sigmasys"],prms["sigmaobs"])
    lines(tmp$t, tmp$y, type='l', col=i)
    return(tmp)
})
## Simulate with ctsmr again with population values
fit$xm["X0"] <- prms["X0"]
fit$xm["a"] <- -1+prms["a"]
fit$xm["b"] <- prms["b"]
fit$xm["p11"] <- log(prms["sigmasys"])
fit$xm["e11"] <- log(prms["sigmaobs"]^2)
plot(D$t, D$y, type='l', lwd=2, ylim=ylim)
LsimCtsmrEstimated <- lapply(1:nsim, function(i){
    tmp <- stochastic_simulate1_ctsmr(fit, newdata=D, nsub=1, firstorder=TRUE, vx0=matrix(0))
    lines(tmp[[1]]$t, tmp[[1]]$output$y, type='l', col=i)
    return(tmp)
})

